#![cfg(feature = "server")]
use std::{error::Error, net::SocketAddr, time::Duration};

use anyhow::anyhow;
use libpt::{
    bintols::display::humanbytes,
    log::{debug, error, info, trace, warn},
};
use threadpool::ThreadPool;
use tokio::{
    io::{AsyncReadExt, BufReader},
    net::{TcpListener, TcpStream},
    time::timeout,
};

use crate::common::conf::Config;

pub mod errors;
use errors::*;

pub struct Server {
    cfg: Config,
    pool: ThreadPool,
    pub timeout: Option<Duration>,
    server: TcpListener,
}

impl Server {
    pub async fn build(cfg: Config) -> anyhow::Result<Self> {
        let server = TcpListener::bind(cfg.addr).await?;
        let pool = ThreadPool::new(cfg.threads);
        let timeout = Some(Duration::from_secs(5));
        Ok(Server {
            cfg,
            pool,
            timeout,
            server,
        })
    }
    pub async fn run(self) -> anyhow::Result<()> {
        loop {
            let (stream, addr) = match self.server.accept().await {
                Ok(s) => s,
                Err(err) => {
                    warn!("could not accept stream: {err:?}");
                    continue;
                }
            };
            match self.handle_stream(stream).await {
                Ok(_) => (),
                Err(err) => {
                    match err {
                        ServerError::Timeout(_) => {
                            info!("stream {:?} timed out", addr)
                        }
                        _ => {
                            warn!("error while handling stream: {:?}", err)
                        }
                    }
                    continue;
                }
            };
        }
    }

    async fn handle_stream(&self, stream: TcpStream) -> Result<()> {
        const BUF_SIZE: usize = 1024;
        let addr = match stream.peer_addr() {
            Ok(a) => a,
            Err(err) => {
                debug!("could not get peer address: {:?}", err);
                return Err(err.into());
            }
        };
        info!("new peer: {:?}", addr);
        let mut buf = [0u8; BUF_SIZE];
        let mut reader = BufReader::new(stream);
        let mut len;
        loop {
            trace!("reading anew");
            len = self.read(&mut reader, &mut buf).await?;
            trace!("did read");
            if len == 0 {
                trace!("len is apperently 0: {len:?}");
                break;
            } else {
                info!(
                    "< {:?}\t({})\n{}",
                    addr,
                    humanbytes(len),
                    self.decode(&buf)?
                );
                buf = [0; BUF_SIZE];
            }
        }
        info!("disconnected peer: {:?}", addr);
        Ok(())
    }

    #[inline]
    fn decode(&self, buf: &[u8]) -> Result<String> {
        Ok(std::str::from_utf8(buf)?.to_string())
    }

    async fn read(&self, reader: &mut BufReader<TcpStream>, buf: &mut [u8]) -> Result<usize> {
        let mut len: usize;
        let mut failsafe: u8 = 0;
        while failsafe < 5 {
            trace!("loop");
            len = match timeout(self.cfg.timeout, reader.read(buf)).await {
                Ok(inner) => {
                    trace!("Read for len: {inner:?}");
                    match inner {
                        Ok(len) => len,
                        Err(err) => {
                            error!("read failed: {err:?}");
                            return Err(ServerError::RdFail);
                        }
                    }
                }
                Err(err) => {
                    return Err(err.into());
                }
            };
            if self.should_end(&buf) {
                return Ok(len);
            } else {
                // ?????????
                failsafe += 1;
            }
        }
        // TODO: remove the failsafe when this stuff works
        Err(anyhow!("read too often, so the failsafe activated").into())
    }

    fn should_end(&self, buf: &[u8]) -> bool {
        let mut lb: u8 = buf[0];
        let mut iter = buf.iter().skip(1).peekable();
        while let Some(b) = iter.next() {
            if lb != 0 && *b == 0 && **iter.peek().unwrap() == 0 {
                return true;
            }
            lb = *b;
        }
        return false;
    }
}
